#include "cui.h"
#include "focus.h"
#include "rect.h"
#include "esc.h"
#include "key.h"
#include "handler.h"
#include "panel.h"
#include "label.h"
#include "button.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

typedef struct cui_main_stack *cui_main_stack;

struct cui_main_stack{
	cui focus;
	int running;
	cui_main_stack prev;
};

static cui_main_stack main_stack = NULL;

cui
cui_focus_get(void)
{
	return main_stack ? main_stack->focus : NULL;
}

void
cui_focus_set(cui obj)
{
	if(main_stack) main_stack->focus = obj;
}

int
cui_running_get(void)
{
	return main_stack ? main_stack->running : FALSE;
}

void
cui_running_set(int v)
{
	if(main_stack) main_stack->running = v;
}

cui
cui_alloc(int size)
{
	cui obj;

	if((obj = malloc(size)) == NULL) ERR("No Mem");
	return obj;
}

void
cui_base_init(cui obj, cui parent, int x, int y, int w, int h)
{
	obj->parent = parent;
	obj->children = NULL;
	obj->next = NULL;
	obj->x = x;
	obj->y = y;
	obj->w = w;
	obj->h = h;

	if(parent) cui_base_child_add(parent, obj);
	obj->handler_list = NULL;
	obj->flags = 0;
}

void
cui_base_child_add(cui obj, cui child)
{
	child->next = obj->children;
	obj->children = child;
}

cui
cui_base_new(cui parent, int x, int y, int w, int h)
{
	cui obj = cui_alloc(sizeof(struct cui_base));
	cui_base_init(obj, parent, x, y, w, h);
	return obj;
}

void
cui_draw_str(cui obj, int x, int y, char *s, int attr)
{
	struct cui_rect t1, t2;
	cui_rect r1 = &t1, r2 = &t2;
	int n = strlen(s);
	char *s2 = NULL;

	if(!cui_visible_rect(obj, r1)) return;
	cui_rect_init(r2, x, y, n, 1);
	if(!cui_rect_and(r1, r2, r2)) return; /* r1 and r2 --> r2 */

	s += r2->x - x;
	if(r2->x + r2->w < x + n){
		n = r2->w;
		if((s2 = malloc(n + 1)) == NULL) ERR("No Mem");
		memcpy(s2, s, n);
		s2[n] = '\0';
		s = s2;
	}
	x = r2->x;
	cui_esc_loc(cui_gx(obj) + x, cui_gy(obj) + y);
	if(attr & CUI_ATTR_ULINE) cui_esc_attr(CUI_ESC_ULINE);
	if(attr & CUI_ATTR_REVERSE) cui_esc_attr(CUI_ESC_REVERSE);
	printf("%s", s);
	if(attr != CUI_ATTR_NORMAL) cui_esc_attr(CUI_ESC_NORMAL);
	fflush(stdout);
	if(s2) free(s2);
}

void
cui_fill_rect(cui obj, int x, int y, int w, int h, char *s, int attr)
{
	int ix, iy;
	int n = strlen(s);

	for(iy=0; iy<h; iy++){
		for(ix=0; ix<w; ix+=n){
			cui_draw_str(obj, x+ix, y+iy, s, attr);
		}
	}
}

void
cui_clear_rect(cui obj, int x, int y, int w, int h)
{
	cui_fill_rect(obj, x, y, w, h, " ", CUI_ATTR_NORMAL);
}

void
cui_clear(cui obj)
{
	cui_clear_rect(obj, 0, 0, obj->w, obj->h);
}

int
cui_gx(cui obj)
{
	return obj == NULL ? 0 : cui_gx(obj->parent) + obj->x;
}

int
cui_gy(cui obj)
{
	return obj == NULL ? 0 : cui_gy(obj->parent) + obj->y;
}

int
cui_glen(cui o1, cui o2)
{
	int dx, dy;

	dx = (cui_gx(o1) + o1->w / 2) - (cui_gx(o2) + o2->w / 2);
	dy = (cui_gy(o1) + o1->h / 2) - (cui_gy(o2) + o2->h / 2);
	return dx * dx + dy * dy;
}

void
cui_draw(cui obj)
{
	int val;
	
	if(obj == NULL) return;
	if(!cui_is_visible(obj)) return;
	val = (obj == cui_focus_get()) ? CUI_DRAW_FOCUS : CUI_DRAW_NORMAL;
	cui_handler_call(obj, obj->handler_list, CUI_EVT_DRAW, val);
	cui_draw_chain(obj->children);
}

void
cui_draw_chain(cui obj)
{
	if(obj == NULL) return;
	cui_draw_chain(obj->next);
	cui_draw(obj);
}

void
cui_hide(cui obj)
{
	obj->flags |= CUI_FLG_HIDE;
	cui_clear(obj);
}

void
cui_show(cui obj)
{
	obj->flags &= ~CUI_FLG_HIDE;
	cui_draw(obj);
}

int
cui_is_visible(cui obj)
{
	if(obj == NULL) return TRUE; /* root */
	return !(obj->flags & CUI_FLG_HIDE) && cui_is_visible(obj->parent);
}

int
cui_visible_rect(cui obj, cui_rect r)
{
	struct cui_rect tmp;
	cui_rect t = &tmp;

	cui_rect_init(r, 0, 0, obj->w, obj->h);
	if(obj->parent == NULL) return !cui_rect_chk_empty(r); /* empty --> FALSE */
	if(!cui_visible_rect(obj->parent, t)) return FALSE;
	t->x -= obj->x;
	t->y -= obj->y;
	return cui_rect_and(r, t, r); /* r and t --> r */
}

void
cui_xywh_set(cui obj, int x, int y, int w, int h)
{
	int visi = cui_is_visible(obj);
	if(visi) cui_hide(obj);
	obj->x = x;
	obj->y = y;
	obj->w = w;
	obj->h = h;
	if(visi) cui_show(obj);
}

void
cui_x_set(cui obj, int x)
{
	cui_xywh_set(obj, x, obj->y, obj->w, obj->h);
}

void
cui_w_set(cui obj, int w)
{
	cui_xywh_set(obj, obj->x, obj->y, w, obj->h);
}

void
cui_quit(void)
{
	cui_running_set(FALSE);
}

void
cui_main(cui top_obj, cui init_focus)
{
	struct cui_main_stack stk;
	cui_main_stack sp = (cui_main_stack)&stk;

	int key, val;
	cui focus, obj;

	sp->focus = init_focus;
	sp->running = TRUE;
	sp->prev = main_stack;
	main_stack = sp;

	cui_esc_enter();
	cui_key_enter();

	cui_clear(top_obj);
	cui_draw(top_obj);
	while(cui_running_get()){
		if((key = cui_key_get2()) == 0) continue;
		focus = cui_focus_get();
		if(focus){
			if(cui_handler_call(focus, focus->handler_list, CUI_EVT_KEY, key)) continue;
		}
		obj = cui_focus_move_judge(top_obj, key, NULL, &val);
		if(obj){
			cui_focus_set(obj);
			cui_draw(focus);
			cui_draw(obj);
		}
	}
	cui_key_exit();
	cui_esc_exit();

	main_stack = sp->prev;
}

void
cui_del_from(cui obj, cui *from)
{
	if(*from == NULL) return;
	if(*from == obj){
		*from = obj->next;
		return;
	}
	cui_del_from(obj, &(*from)->next);
}

void
cui_del(cui obj)
{
	cui_del_from(obj, &obj->parent->children);
}

void
cui_free_chain(cui obj)
{
	if(obj == NULL) return;
	cui_free_chain(obj->next);
	cui_free(obj);
}

void
cui_free(cui obj)
{
	if(obj == NULL) return;
	cui_free_chain(obj->children);
	cui_handler_list_free(obj->handler_list);
	free(obj);
}

/* EOF */
